@@ -47,7 +47,6 @@ class CreateAgent:
         save_code (bool): If True, extracts and saves code snippets from the response.
         scan_dirs (bool): If True, recursively scans directories found in prompt for existing files, extracts contents, and adds to prompt.
         logging (bool): If True, logs the session to a file.
-        summary (bool): If True, summarizes the current conversation context to reference later.
         api_key (str): The API key for OpenAI. Defaults to system environment variable.
         seed (int or str): Seed for reproducibility. Can be an integer or a string converted to binary.
         iterations (int): Number of response iterations for refining or condensing outputs.
@@ -60,14 +59,17 @@ class CreateAgent:
         glyph (bool): If True, restructures queries with representative/associative glyphs and logic flow
         temperature (float): Range from 0.0 to 2.0, lower values increase randomness, and higher values increase randomness.
         top_p (float): Range from 0.0 to 2.0, lower values increase determinism, and higher values increase determinism.
+        thread_limit (int): Maximum number of messages to a single thread before summarizing content and passing to new instance
+        last_message (str): Last returned system message
 
     Current role shortcuts:
         assistant: Standard personal assistant with improved ability to help with tasks
-        analyst: Expertise in bioinformatics and systems biology. Knowledgeable in commonly used computational biology platforms.
         developer: Generates complete, functional application code based on user requirements, ensuring clarity and structure.
+        prompt: Specializes in analyzing and refining AI prompts to enhance clarity, specificity, and effectiveness without executing tasks.
         refactor: Senior full stack developer with emphases in correct syntax and documentation
         tester: Quality assurance tester with experience in software testing and debugging, generates high-quality unit tests
-        dataviz: Create clear, insightful data visualizations and provide analysis using structured formats, focusing solely on visualization requests and recommendations.
+        analyst: For structured data analysis tasks, adhering to strict validation rules, a detailed workflow, and professional reporting
+        visualize: Create clear, insightful data visualizations and provide analysis using structured formats, focusing solely on visualization requests and recommendations.
         writer: Writing assistant to help with generating science & technology related content
         editor: Text editing assistant to help with clarity and brevity
         artist: Creates an images described by the prompt, default style leans toward illustrations
@@ -76,8 +78,9 @@ class CreateAgent:
     Methods:
         __init__: Initializes the handler with default or provided values.
         request: Submits a query to the OpenAI API and processes the response.
-        new_thread: Start a new thread with only the current agent.
-        _save_chat_transcript: Saves the conversation transcript to a file.
+        report_status: Reports current attributes and status of agent
+        start_new_thread: Start a new thread with only the current agent.
+        summarize_current_thread: Summarize current conversation history for future context parsing.
         _extract_and_save_code: Extracts code snippets from the response and saves them to files.
         _setup_logging: Prepares logging setup.
         _prepare_query_text: Prepares the query, including prompt modifications and image handling.
@@ -105,7 +108,6 @@ class CreateAgent:
         chain_of_thought=False,
         save_code=False,
         scan_dirs=False,
-        summary=False,
         model="gpt-4o-mini",
         role="assistant",
         seed="t634e``R75T86979UYIUHGVCXZ",
@@ -114,17 +116,12 @@ class CreateAgent:
         top_p=1.0,
         dimensions="NA",
         quality="NA",
-        mode='normal',
-    ):
+        stage='normal',
+        thread_limit=10):
         """
         Initialize the handler with default or provided values.
         """
-        # Globals
-        self.client = client
-        self.thread = thread
-        self.api_key = api_key
-
-        # Booleans
+        self.logging = logging
         self.verbose = verbose
         self.silent = silent
         self.refine_prompt = refine
@@ -132,14 +129,33 @@ class CreateAgent:
         self.chain_of_thought = chain_of_thought
         self.save_code = save_code
         self.scan_dirs = scan_dirs
+        self.model = model
+        self.role = role
+        self.seed = seed
         self.iterations = iterations
-        self.summary = summary
+        self.temperature = temperature
+        self.top_p = top_p
+        self.dimensions = dimensions
+        self.quality = quality
+        self.stage = stage
+        self.current_thread_limit = thread_limit
 
-        # Additional hyperparams
-        self.mode = mode if mode == 'refine_only' else 'normal'
+        # Check user input types
+        self._validate_types()
+
+        # API data
+        self.api_key = api_key
+        self.client = client
+        self.thread = thread
+        self.thread.current_thread_calls = 0
+        self.thread.current_thread_limit = self.current_thread_limit
+        self.total_cost = 0.0
+        
+        # Validdate specific hyperparams
+        self.stage = self.stage if self.stage == 'refine_only' else 'normal'
         self.tokens = {"prompt": 0, "completion": 0}
-        self.seed = seed if isinstance(seed, int) else self._string_to_binary(seed)
-        self._validate_probability_params(temperature, top_p)
+        self.seed = self.seed if isinstance(self.seed, int) else self._string_to_binary(self.seed)
+        self.temperature, self.top_p = self._validate_probability_params(self.temperature, self.top_p)
         
         # Validate user inputs
         self._prepare_system_role(role)
@@ -151,14 +167,57 @@ class CreateAgent:
         # Initialize reporting and related vars
         self.timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
         self.prefix = f"{self.label}.{self.model.replace('-', '_')}.{self.timestamp}"        
-        self._setup_logging(logging)
-        self._log_and_print(self._generate_status(), self.verbose, self.logging)
+        if self.logging: self._setup_logging()
+        self._log_and_print(self.report_status(), False, self.logging)
 
-    def _setup_logging(self, log):
+    def _validate_types(self):
+        """
+        Validates the types of the instance attributes for CreateAgent.
+
+        Raises:
+            TypeError: If any attribute has an incorrect type.
+            ValueError: If any integer attribute is not positive.
+        """
+        expected_types = {
+            'logging': bool,
+            'verbose': bool,
+            'silent': bool,
+            'refine_prompt': bool,
+            'glyph_prompt': bool,
+            'chain_of_thought': bool,
+            'save_code': bool,
+            'scan_dirs': bool,
+            'model': str,
+            'role': str,
+            'seed': (int, str),  # seed can be either int or str
+            'iterations': int,
+            'temperature': float,
+            'top_p': float,
+            'dimensions': str,
+            'quality': str,
+            'stage': str,
+            'current_thread_limit': int
+        }
+
+        for attr_name, expected_type in expected_types.items():
+            value = getattr(self, attr_name, None)  # Get the attribute value from self
+            if isinstance(expected_type, tuple):
+                # Check if value matches any expected type in the tuple
+                if not isinstance(value, expected_type):
+                    raise TypeError(f"Expected type for {attr_name} is {expected_type}, got {type(value).__name__}")
+            else:
+                # Check if value matches the expected type
+                if not isinstance(value, expected_type):
+                    raise TypeError(f"Expected type for {attr_name} is {expected_type}, got {type(value).__name__}")
+            
+            # Check if integer-type values are positive
+            if expected_type == int and value <= 0:
+                raise ValueError(f"{attr_name} must be a positive integer, got {value}")
+
+    def _setup_logging(self):
         """
         Prepare logging setup.
         """
-        self.logging = log
         self.log_text = []
         self.log_file = f"logs/{self.prefix}.transcript.log"
         os.makedirs("logs", exist_ok=True)
@@ -167,18 +226,17 @@ class CreateAgent:
 
     def _validate_probability_params(self, temp, topp):
         """Ensure temperature and top_p are valid"""
-        self.temperature = temp
-        self.top_p = topp
-
         # Acceptable ranges
-        if self.temperature < 0.0 or self.temperature > 2.0:
-            self.temperature = 0.7
-        if self.top_p < 0.0 or self.top_p > 2.0:
-            self.top_p = 1.0
+        if temp < 0.0 or temp > 2.0:
+            temp = 0.7
+        if topp < 0.0 or topp > 2.0:
+            topp = 1.0
 
         # Only one variable is changed at a time
-        if self.temperature != 0.7 and self.top_p != 1.0:
-            self.top_p = 1.0
+        if temp != 0.7 and topp != 1.0:
+            topp = 1.0
+
+        return temp, topp
 
     def _prepare_query_text(self, prompt_text):
         """
@@ -220,7 +278,7 @@ class CreateAgent:
             self.label = "default"
             self.role_name = "Default User"
         else:
-            self.role = role
+            self.role = input_role
             self.label = "custom"
             self.role_name = "User-defined custom role"
 
@@ -245,35 +303,53 @@ class CreateAgent:
         self.quality = "hd" if quality.lower() in {"h", "hd", "high", "higher", "highest"} else "standard"
         self.quality = "hd" if self.label == "photographer" else self.quality # Check for photo role
         
-    def _generate_status(self):
+    def report_status(self):
         """Generate status message."""
         status = f"""
 Agent parameters:
     Model: {self.model}
     Role: {self.role_name}
+    
     Chain-of-thought: {self.chain_of_thought}
     Prompt refinement: {self.refine_prompt}
     Associative glyphs: {self.glyph_prompt}
     Response iterations: {self.iterations}
     Subdirectory scanning: {self.scan_dirs}
-    Time stamp: {self.timestamp}
-    Seed: {self.seed}
     Text logging: {self.logging}
     Verbose StdOut: {self.verbose}
-    Snippet logging: {self.save_code}
+    Code snippet detection: {self.save_code}
+    Time stamp: {self.timestamp}
+    Assistant ID: {self.agent}
+    Thread ID: {self.thread.id}
+    Seed: {self.seed}
+    Requests in current thread: {self.thread.current_thread_calls}
+    Current total cost: ${round(self.total_cost, 5)}
     """
         if "dall-e" in self.model:
             status += f"""Image dimensions: {self.dimensions}
     Image quality: {self.quality}
-    """
-        return status
+"""
+        print(status)
 
-    def new_thread(self):
-        """Start a new thread with only the current agent."""
+    def start_new_thread(self, context=None):
+        """Start a new thread with only the current agent and adds previous context if needed."""
         self.thread = self.client.beta.threads.create()
+        self.thread.current_thread_calls = 0
+        self.thread.current_thread_limit = self.current_thread_limit
 
-    def request(self, prompt, thread=None):
+        # Add previous context
+        if context:
+            previous_context = client.beta.threads.messages.create(
+                thread_id=self.thread.id, role="user", content=context)
+
+    def request(self, prompt=''):
         """Submits the query to OpenAIs API and processes the response."""
+        # Checks for last system response is not prompt provided
+        if prompt == '':
+            try:
+                prompt = self.last_message
+            except Exception as e:
+                raise ValueError(f"No existing messages found in thread: {e}")
 
         # Update user prompt 
         self._prepare_query_text(prompt)
@@ -281,8 +357,9 @@ Agent parameters:
             f"\n{self.model} processing updated conversation thread...\n",
                 True, self.logging)
 
-        if self.mode != "refine_only":
+        if self.stage != "refine_only":
             if "dall-e" not in self.model:
+                self.thread.current_thread_calls += 1
                 self._handle_text_request()
             else:
                 self._handle_image_request()
@@ -290,9 +367,11 @@ Agent parameters:
         token_report = self._gen_token_report()
         self._log_and_print(token_report, self.verbose, self.logging)
 
-        # Save chat logs to file
-        if self.logging:
-            self._save_chat_transcript()
+        # Check current scope thread
+        if self.thread.current_thread_calls >= self.thread.current_thread_limit:
+            self._log_and_print(f"\nReached end of current thread limit.\n", self.verbose, False)
+            summary = self.summarize_current_thread()
+            self.start_new_thread("The following is a summary of a ongoing conversation with a user and an AI assistant:\n" + summary)
 
     def _init_chat_completion(self, prompt, model='gpt-4o-mini', role='user', iters=1, seed=42, temp=0.7, top_p=1.0):
         """Initialize and submit a single chat completion request"""
@@ -304,21 +383,36 @@ Agent parameters:
 
         return completion
 
-    def _generate_context_summary(self):
+    def summarize_current_thread(self):
         """Summarize current conversation history for future context parsing."""
-        self._log_and_print(f"\ngpt-4o-mini summarizing current conversation...\n", True, False)
+        self._log_and_print(f"\ngpt-4o-mini summarizing current thread...\n", self.verbose, False)
 
-        summarized = self._init_chat_completion(self, 
-            prompt=f"{modifierDict['summarize']}\n\n{'\n'.join(self.log_text)}", 
-            iters=self.iterations, seed=self.seed)
+        # Get all thread messages
+        all_messages = self._get_thread_messages()
+
+        # Generate concise summary
+        summary_prompt = modifierDict['summarize'] + "\n\n" + all_messages
+        summarized = self._init_chat_completion(prompt=summary_prompt, iters=self.iterations, seed=self.seed)
         self._update_token_count(summarized)
-        self.current_context = summarized.choices[0].message.content.strip()
+
+        return summarized.choices[0].message.content.strip()
+
+    def _get_thread_messages(self, join_messages=True):
+        """Fetches all messages from a thread in order and returns them as a list."""
+        messages = self.client.beta.threads.messages.list(thread_id=self.thread.id)
+        sorted_messages = sorted(messages.data, key=lambda msg: msg.created_at)
+        conversation = "\n".join(sorted_messages)
+
+        if join_messages == True:
+            conversation = "\n".join(conversation)
+
+        return conversation
 
     def _handle_text_request(self):
         """Processes text-based responses from OpenAIs chat models."""
-        self.message = self._run_thread_request()
+        self.last_message = self._run_thread_request()
         self._update_token_count(self.run_status)
-        self._log_and_print(self.message, True, self.logging)
+        self._log_and_print(self.last_message, True, self.logging)
 
         # Extract code snippets
         code_snippets = self._extract_code_snippets()
@@ -334,9 +428,15 @@ Agent parameters:
 
             self._log_and_print(reportStr, True, self.logging)
 
-        # Summarize current context
-        if self.summary == True:
-            self._generate_context_summary(self)
+        # Check URL annotations - inactive for now
+        #existing, not_existing = self._check_response_urls()
+        #if len(not_existing) >= 1 or len(existing) >= 1:
+        #    reportStr = "\nURL citations detecting in system message\n"
+        #    if len(existing) >= 1:
+        #        reportStr += 'Found:\n\t' '\n\t'.join(existing) + '\n'
+        #    if len(not_existing) >= 1:
+        #        reportStr += 'NOT found:\n\t' '\n\t'.join(not_existing) + '\n'
+        #    self._log_and_print(reportStr, self.verbose, self.logging)
 
     def _write_script(self, content, file_name, outDir="code", lang=None):
         """Writes code to a file."""
@@ -369,19 +469,18 @@ Agent parameters:
         with open(image_file, "wb") as outFile:
             outFile.write(image_data)
 
-        self.message = (
+        self.last_message = (
             "\nRevised image prompt:\n"
             + response.data[0].revised_prompt
             + "\nGenerated image saved to:\n"
             + image_file
         )
-        self._log_and_print(self.message, True, self.logging)
+        self._log_and_print(self.last_message, True, self.logging)
 
     def _gen_token_report(self):
         """Generates report string for overall cost of the query."""
         prompt_cost, completion_cost = "Unknown model rate", "Unknown model rate"
         total_tokens = self.tokens["prompt"] + self.tokens["completion"]
-        self.total_cost = "Unknown model rate"
         rates = {
             "gpt-4o": (2.5, 10),
             "gpt-4o-mini": (0.150, 0.600),
@@ -396,34 +495,26 @@ Agent parameters:
             completion_cost = self._calculate_cost(
                 self.tokens["completion"], completion_rate
             )
-            self.total_cost = round(prompt_cost + completion_cost, 5)
+            self.total_cost += round(prompt_cost + completion_cost, 5)
         return (
-            f"\nTotal tokens generated: {total_tokens}  (${self.total_cost})"
-            f"\n    Prompt (i.e. input): {self.tokens['prompt']}  (${prompt_cost})"
-            f"\n    Completion (i.e. output): {self.tokens['completion']}  (${completion_cost})"
-        )
-
-    def _save_chat_transcript(self):
-        """Saves the current response text to a file if specified."""
-        with open(self.log_file, "a", encoding="utf-8") as f:
-            f.write("\n".join(self.log_text))
-        self._log_and_print(
-            f"\nSaving conversation transcript text to: {self.log_file}\n",
-            self.verbose,
-            self.logging,
+            f"\nCurrent total tokens generated by this agent: {total_tokens}  (${self.total_cost})"
+            f"\n - prompt tokens (i.e. input): {self.tokens['prompt']}  (${prompt_cost})"
+            f"\n - completion tokens (i.e. output): {self.tokens['completion']}  (${completion_cost})"
         )
 
     def _condense_iterations(self, api_response):
         """Condenses multiple API responses into a single coherent response."""
         api_responses = [r.message.content.strip() for r in api_response.choices]
-        api_responses = self._gen_iteration_str(api_responses)
+        api_responses =  "\n\n".join(
+            ["\n".join([f"Iteration: {i + 1}", api_responses[i]])
+            for i in range(len(api_responses))])
 
         self._log_and_print(
             f"\ngpt-4o-mini condensing response iterations...", self.verbose, self.logging
         )
-        condensed = self._init_chat_completion(self, 
-            prompt=f"{modifierDict['condense']}\n\n{api_responses}", 
-            role=self.role, iters=self.iterations, seed=self.seed)
+        condensed = self._init_chat_completion( 
+            prompt= modifierDict['condense'] + "\n\n" + api_responses, 
+            iters=self.iterations, seed=self.seed)
         self._update_token_count(condensed)
         message = condensed.choices[0].message.content.strip()
         self._log_and_print(
@@ -460,8 +551,9 @@ Agent parameters:
         if self.glyph_prompt == True:
             updated_prompt += modifierDict["glyph"]
 
-        refined = self._init_chat_completion(self, 
+        refined = self._init_chat_completion(
             prompt=updated_prompt, 
+            role=self.role,
             seed=self.seed, 
             iters=self.iterations,
             temp=self.temperature, 
@@ -474,7 +566,7 @@ Agent parameters:
             new_prompt = refined.choices[0].message.content.strip()
 
         self._log_and_print(
-            f"\nRefined query prompt:\n{new_prompt}", self.verbose, self.logging)
+            f"Refined query prompt:\n{new_prompt}", self.verbose, self.logging)
 
         return new_prompt
 
@@ -489,7 +581,8 @@ Agent parameters:
             if verb == True and self.silent == False:
                 print(message)
             if log == True:
-                self.log_text.append(message)
+                with open(self.log_file, "a", encoding="utf-8") as f:
+                    f.write(message + "\n")
 
     @staticmethod
     def _calculate_cost(tokens, perM, dec=5):
@@ -559,7 +652,7 @@ Agent parameters:
         # Regular expression to match code blocks enclosed in triple backticks, including the language tag
         code_snippets = defaultdict(str)
         code_pattern = re.compile(r"```(\w+)\n(.*?)```", re.DOTALL)
-        snippets = code_pattern.findall(self.message)
+        snippets = code_pattern.findall(self.last_message)
         for lang, code in snippets:
             code_snippets[lang] += code.strip()
 
@@ -627,11 +720,12 @@ Agent parameters:
             New assistant assistant class instance
         """
         try:
-            self.agent = self.client.beta.assistants.create(
+            agent = self.client.beta.assistants.create(
                 name=self.role_name,
                 instructions=self.role,
                 model=self.model,
                 tools=[{"type": "code_interpreter"}] if interpreter == True else [])
+            self.agent = agent.id
         except Exception as e:
             raise RuntimeError(f"Failed to create assistant: {e}")
 
@@ -656,7 +750,7 @@ Agent parameters:
         # Run the assistant on the thread
         current_run = self.client.beta.threads.runs.create(
             thread_id=self.thread.id,
-            assistant_id=self.agent.id)
+            assistant_id=self.agent)
 
         # Wait for completion and retrieve responses
         while True:
@@ -664,7 +758,7 @@ Agent parameters:
             if self.run_status.status in ["completed", "failed"]:
                 break
             else:
-                time.sleep(2)  # Wait before polling again
+                time.sleep(1)  # Wait before polling again
 
         if self.run_status.status == "completed":
             messages = self.client.beta.threads.messages.list(thread_id=self.thread.id)
@@ -674,3 +768,42 @@ Agent parameters:
                 raise ValueError("No messages found in the thread.")
         else:
             raise ValueError("Assistant failed to generate a response.")
+
+    def _check_response_urls(self):
+        """
+        Extracts all URLs from the given text using regex,checks the existence of 
+        each URL, and returns lists of existing and non-existing URLs.
+        
+        Args:
+            text (str): The input text containing potential URLs.
+
+        Returns:
+            Tuple[List[str], List[str]]: A tuple containing two lists - 
+                                           the first list for existing URLs,
+                                           and the second for non-existing URLs.
+        """
+        # Define a regex pattern for URL extraction.
+        url_pattern = r'https?://[^\s]+|ftp://[^\s]+'
+        urls = re.findall(url_pattern, self.last_message)
+        
+        # Check if identified URLs are real
+        existing_urls = non_existing_urls = []
+        for url in urls:
+            try:
+                # Execute a curl command to check URL existence
+                response = subprocess.run(['curl', '-Is', url], capture_output=True, text=True)
+                
+                if response.returncode == 0:
+                    # Extract status code
+                    status_line = response.stdout.splitlines()[0]
+                    status_code = status_line.split()[1]
+                    if status_code.startswith('2'):  # Status codes 2xx indicate success
+                        existing_urls.append(url)
+                    else:
+                        non_existing_urls.append(url)
+                else:
+                    non_existing_urls.append(url)  # If curl fails, consider the URL as non-existing
+            except Exception as e:
+                non_existing_urls.append(url)  # Append URL in case of any exception
+
+        return existing_urls, non_existing_urls
